#include "js-gtk.h"

#define BUTTON_CTOR(type) \
static JSBool type##_ctor(JSContext *cx, JSObject *obj,                      \
			  uintN argc, jsval *argv, jsval *rval) {            \
    GtkWidget *type;                                                         \
    if (!JS_IsConstructing(cx))                                              \
	return JS_FALSE;                                                     \
    if (argc) {                                                              \
	char *label = JS_GetStringBytes(JS_ValueToString(cx, argv[0]));      \
	type = gtk_##type##_new_with_label(label);                           \
    } else {                                                                 \
	type = gtk_##type##_new();                                           \
    }                                                                        \
    if (!type)                                                               \
	return JS_FALSE;                                                     \
    JS_SetPrivate(cx, obj, type);                                            \
    jsgtkobj_registerWrapper(cx, GTK_OBJECT(type), obj);                     \
    return JS_TRUE;                                                          \
}

/* check button */

BUTTON_CTOR(check_button);
JSGTK_CLASS(CheckButton);

static JSBool jsgtkcheckbutton_init(JSContext *cx, JSObject *global,
				    JSObject *gtkproto) {
    JSObject *checkbuttonProto;

    checkbuttonProto = JS_InitClass(cx, global, gtkproto,
				    &jsgtk_CheckButtonClass,
				    check_button_ctor, 0,
				    NULL, NULL,
				    NULL, NULL);
    if (!checkbuttonProto)
	/* 	!jsgtkradiobutton_init(cx, global, checkbuttonProto)) */
	return JS_FALSE;
    
    return JS_TRUE;
}

/* toggle button */

#define INT_FUNC(funcname) \
  INT_WRAPPER_FUNC(toggle_button,ToggleButton,TOGGLE_BUTTON,funcname)
#define SIMPLE_FUNC(funcname) \
  SIMPLE_WRAPPER_FUNC(toggle_button,ToggleButton,TOGGLE_BUTTON,funcname)

INT_FUNC(set_state)
SIMPLE_FUNC(toggled);

#undef INT_FUNC
#undef SIMPLE_FUNC

JSGTK_CLASS(ToggleButton);

static JSFunctionSpec toggle_button_methods[] = {
    {"set_state",	toggle_button_set_state,	1},
    {"toggled",		toggle_button_toggled,		0},
    {0}
};

static JSBool toggle_set_active(JSContext *cx, JSObject *obj,
				jsval id, jsval *vp) {
    GtkToggleButton *toggle;
    JSBool newstate;

    if (JSVAL_TO_INT(id) != 42 ||
	!JSVAL_IS_BOOLEAN(*vp))
	return JS_FALSE;

    GET_GTK(toggle,cx,obj,TOGGLE_BUTTON);
    if (!JS_ValueToBoolean(cx, *vp, &newstate))
	return JS_FALSE;
    gtk_toggle_button_set_state(toggle, (gint)newstate);
    return JS_TRUE;
}

static JSBool toggle_get_active(JSContext *cx, JSObject *obj,
				jsval id, jsval *vp) {
    GtkToggleButton *toggle;
    
    if (JSVAL_TO_INT(id) != 42)
	return JS_FALSE;

    GET_GTK(toggle,cx,obj,TOGGLE_BUTTON);
    *vp = toggle->active ? JSVAL_TRUE : JSVAL_FALSE;
    return JS_TRUE;
}

static JSPropertySpec toggle_button_props[] = {
    {"active",	42,	JSPROP_ENUMERATE|JSPROP_PERMANENT,
     toggle_get_active, toggle_set_active},
    {0}
};

BUTTON_CTOR(toggle_button);

static JSBool jsgtktogglebutton_init(JSContext *cx, JSObject *global,
				     JSObject *gtkproto) {
    JSObject *togglebuttonProto;
    
    togglebuttonProto = JS_InitClass(cx, global, gtkproto,
				     &jsgtk_ToggleButtonClass,
				     toggle_button_ctor, 0,
				     toggle_button_props,
				     toggle_button_methods,
				     NULL, NULL);
    if (!togglebuttonProto ||
	!jsgtkcheckbutton_init(cx, global, togglebuttonProto))
	return JS_FALSE;

    return JS_TRUE;
}

/* simple button */

#define SIMPLE_FUNC(funcname) \
  SIMPLE_WRAPPER_FUNC(button,Button,BUTTON,funcname)

SIMPLE_FUNC(pressed)
SIMPLE_FUNC(released)
SIMPLE_FUNC(clicked)
SIMPLE_FUNC(enter)
SIMPLE_FUNC(leave)

#undef SIMPLE_FUNC

JSGTK_CLASS(Button);

static JSFunctionSpec button_methods[] = {
    {"pressed",		button_pressed,		0},
    {"released",	button_released,	0},
    {"clicked",		button_clicked,		0},
    {"enter",		button_enter,		0},
    {"leave",		button_leave,		0},
    {0}
};

BUTTON_CTOR(button);

JSBool jsgtkbutton_init(JSContext *cx, JSObject *global, JSObject *gtkproto) {
    JSObject *buttonProto;

    buttonProto = JS_InitClass(cx, global, gtkproto, &jsgtk_ButtonClass,
			       button_ctor, 1,
			       NULL, button_methods,
			       NULL, NULL);
    if (!buttonProto ||
	!jsgtktogglebutton_init(cx, global, buttonProto))
	return JS_FALSE;
    return JS_TRUE;
}

