#include "jsapi.h"
#include "js-gtk.h"
#include <assert.h>
#include <stdio.h>

#define STACK_CHUNK_SIZE                8192

static void
init_errorReporter(JSContext *cx, const char *msg, JSErrorReport *report) {
  printf("***** Error: %s (%s:%d)\n       line: %s\n      token: %s\n",
         msg, report ? report->filename : "", 
	 report ? report->lineno : 0,
	 report ? report->linebuf : "",
         report ? report->tokenptr : "");
}

static JSClass global_class = {
  "global", 0,
  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   JS_FinalizeStub
};

/* This is stupid... damned C scoping. */
#define EXT_JSFUNC(func)        static JSBool \
func(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval) { \
  return my_##func(cx, obj, argc, argv, rval); \
						 } \

EXT_JSFUNC(Print);

static JSBool init_gtkmain(JSContext *cx, JSObject *obj,
			   uintN argc, jsval *argv, jsval *rval) {
    gtk_main();
    return JS_TRUE;
}

static JSBool init_gtkmainquit(JSContext *cx, JSObject *obj,
			       uintN argc, jsval *argv, jsval *rval) {
    gtk_main_quit();
    return JS_TRUE;
}

static JSFunctionSpec my_functions[] = {
  {"print",             Print,          	0},
  {"gtk_main",		init_gtkmain,		0},
  {"gtk_main_quit",	init_gtkmainquit,	0},
  { 0 }
};

JSBool init_doInit(JSContext **cx, JSObject **globalObj, int argc, 
		   char **argv) {

  JSTaskState *task;
  task = JS_Init(8L * 1024L * 1024L);
  if (!task)
    return 0;

  *cx = JS_NewContext(task, STACK_CHUNK_SIZE);
  if (!*cx)
      return 0;
  
  *globalObj = JS_NewObject(*cx, &global_class, 0, 0);
  
  if (!JS_InitStandardClasses(*cx, *globalObj))
      return 0;

  if (!JS_DefineFunctions(*cx, *globalObj, my_functions))
      return 0;

  if (!jsgtkobj_init(*cx, *globalObj, argc, argv))
      return 0;
  
  JS_SetErrorReporter(*cx, init_errorReporter);

  return 1;
}
