#include "js-gtk.h"
/* JG 1999 02 27 */
/* #include "prhash.h" */
#include <assert.h>

/* Horizontal box */

JSGTK_CLASS(HBox);

static JSBool hbox_ctor(JSContext *cx, JSObject *obj,
			uintN argc, jsval *argv, jsval *rval) {
    GtkWidget *hbox;
    uint32 arg1;
    
    if (argc < 2 ||
	!JSVAL_IS_BOOLEAN(argv[0]) ||
	!JSVAL_IS_INT(argv[1]) ||
	!JS_IsConstructing(cx)) {
	JS_ReportError(cx, "Bad types for GtkHBox: %s,%s",
		       JS_GetTypeName(cx, JS_TypeOfValue(cx, argv[0])),
		       JS_GetTypeName(cx, JS_TypeOfValue(cx, argv[1])));
	return JS_FALSE;
    }
    
    if (!JS_ValueToInt32(cx, argv[0], &arg1))
	return JS_FALSE;
    hbox = gtk_hbox_new(arg1, JSVAL_TO_INT(argv[1]));
    JS_SetPrivate(cx, obj, hbox);
    jsgtkobj_registerWrapper(cx, GTK_OBJECT(hbox), obj);
    return JS_TRUE;
}

#define hbox_methods NULL

JSGTK_INIT_FUNC(hbox,HBox,2);

/* Vertical box */

JSGTK_CLASS(VBox);

static JSBool vbox_ctor(JSContext *cx, JSObject *obj,
			uintN argc, jsval *argv, jsval *rval) {
    GtkWidget *vbox;
    uint32 arg1;
    
    if (argc < 2 ||
	!JSVAL_IS_BOOLEAN(argv[0]) ||
	!JSVAL_IS_INT(argv[1]) ||
	!JS_IsConstructing(cx)) {
	JS_ReportError(cx, "Bad types for GtkVBox: %s,%s",
		       JS_GetTypeName(cx, JS_TypeOfValue(cx, argv[0])),
		       JS_GetTypeName(cx, JS_TypeOfValue(cx, argv[1])));
	return JS_FALSE;
    }
    
    if (!JS_ValueToInt32(cx, argv[0], &arg1))
	return JS_FALSE;
    vbox = gtk_vbox_new(arg1, JSVAL_TO_INT(argv[1]));
    JS_SetPrivate(cx, obj, vbox);
    jsgtkobj_registerWrapper(cx, GTK_OBJECT(vbox), obj);
    return JS_TRUE;
}

#define vbox_methods NULL
JSGTK_INIT_FUNC(vbox,VBox,2);

/* Base box class */

#define WIDGET_FUNC(funcname) \
  WIDGET_WRAPPER_FUNC(box,Box,BOX,funcname)

WIDGET_FUNC(pack_start_defaults)
WIDGET_FUNC(pack_end_defaults);

#undef WIDGET_FUNC

#define PACK_FUNC(name)\
static JSBool box_pack_##name(JSContext *cx, JSObject *obj,                   \
			    uintN argc, jsval *argv, jsval *rval) {           \
    GtkBox *box;                                                              \
    GtkWidget *widget;                                                        \
    int32 arg1, arg2;	                                                      \
    if (argc < 4 ||                                                           \
	!JSVAL_IS_OBJECT(argv[0]) ||                                          \
	!JSVAL_IS_BOOLEAN(argv[1]) || !JSVAL_IS_BOOLEAN(argv[2]) ||           \
	!JSVAL_IS_INT(argv[3])) {                                             \
        JS_ReportError(cx, "bad types to GtkBox." #name ": %d/%s\n",          \
		       argc, jsgtk_print_types(cx, argv, argc));              \
	return JS_FALSE;                                                      \
    }                                                                         \
    GET_GTK(box,cx,obj,BOX);                                                  \
    GET_GTK(widget,cx,JSVAL_TO_OBJECT(argv[0]),WIDGET);                       \
    if (!JS_ValueToInt32(cx, argv[1], &arg1) ||                               \
	!JS_ValueToInt32(cx, argv[2], &arg2))                                 \
	return JS_FALSE;                                                      \
    gtk_box_pack_##name(box, widget, arg1, arg2, JSVAL_TO_INT(argv[3]));      \
    return JS_TRUE;                                                           \
}

PACK_FUNC(start)
PACK_FUNC(end);

#undef PACK_FUNC

JSGTK_CLASS(Box);

static JSFunctionSpec box_methods[] = {
    {"pack_start_defaults",	box_pack_start_defaults,	1},
    {"pack_end_defaults",	box_pack_end_defaults,		1},
    {"pack_end",		box_pack_end,			4},
    {"pack_start",		box_pack_start,			4},
    {0}
};

JSBool jsgtkhbox_init(JSContext *cx, JSObject *global, JSObject *gtkproto);
JSBool jsgtkvbox_init(JSContext *cx, JSObject *global, JSObject *gtkproto);

JSBool jsgtkbox_init(JSContext *cx, JSObject *global, JSObject *gtkproto) {
    JSObject *boxProto;
    
    boxProto = JS_InitClass(cx, global, gtkproto, &jsgtk_BoxClass,
			    NULL, 0,
			    NULL, box_methods,
			    NULL, NULL);
    if (!boxProto ||
	!jsgtkhbox_init(cx, global, boxProto) ||
	!jsgtkvbox_init(cx, global, boxProto))
	return JS_FALSE;
    
    return JS_TRUE;
}

/* object.c */

struct gtkobj_signalArg {
    JSContext *cx;		/* blah */
    JSObject *obj;		/* closure/function */
};

typedef struct gtkobj_signalArg gtkobj_signalArg;

JSGTK_CLASS(Object);

static GHashTable *wrapperHash;

/* JG 1999 02 27 */
/*static prhashcode StubHash(const void *key) {
    return (prhashcode)key;
}
*/

/* exported functions */
char *jsgtk_print_types(JSContext *cx, jsval *argv, uintN argc) {
    char *output;
    int i;

    output = JS_malloc(cx, argc * 10);
    output[0] = '(';
    for (i = 0; i < argc; i++) {
	if (i)
	    strcat(output, ",");
	strcat(output, JS_GetTypeName(cx, JS_TypeOfValue(cx, argv[i])));
    }
    strcat(output, ")");
    return output;
}

JSBool jsgtkobj_registerWrapper(JSContext *cx, GtkObject *gtkobj,
			      JSObject *obj) {
    gtk_object_ref(gtkobj);
    //   return (JSBool)PR_HashTableAdd(wrapperHash, gtkobj, obj);
    // JG 1999 02 27
    g_hash_table_insert(wrapperHash, gtkobj, obj);
    return JS_TRUE;
}

JSObject *jsgtkobj_findWrapper(JSContext *cx, GtkObject *gtkobj) {
    //    return (JSObject *)PR_HashTableLookup(wrapperHash, gtkobj);
    // JG 1999 02 27
    return (JSObject *) g_hash_table_lookup(wrapperHash, gtkobj);
}

JSBool jsgtkobj_unregisterWrapper(JSContext *cx, GtkObject *gtkobj) {
    gtk_object_unref(gtkobj);
    //return (JSBool)PR_HashTableRemove(wrapperHash, gtkobj);
    // JG 1999 02 27
    g_hash_table_remove(wrapperHash, gtkobj);
    return JS_TRUE;
}

void jsgtkobj_finalize(JSContext *cx, JSObject *obj) {
    GtkObject *gtkobj = JS_GetPrivate(cx, obj);
    /* unregister wrapper */
    assert(GTK_IS_OBJECT(gtkobj));
    jsgtkobj_unregisterWrapper(cx, gtkobj);
}

#define CHECK(Type, TYPE)           \
if (GTK_IS_##TYPE(gtkobj)) {       \
  cls = &jsgtk_##Type##Class;      \
  proto = jsgtk_##Type##Proto;     \
  goto found;                      \
}

JSObject *jsgtkobj_wrapObject(JSContext *cx, GtkObject *gtkobj,
			      JSObject *parent) {
    JSObject *obj, *proto;
    JSClass *cls;   

    /* JG 1999 02 27  */
   /*
     CHECK(hbox,HBox);
     CHECK(vbox,VBox);
   */    
    CHECK(HBox, HBOX);
    CHECK(VBox,VBOX);

    /* XXXshaver check type! */

 found:
    obj = JS_NewObject(cx, cls, proto, parent);
    if (!obj) {
	JS_ReportError(cx, "Can't create wrapper object for %x!\n",
		       gtkobj);
	goto out;
    }
    JS_SetPrivate(cx, obj, gtkobj);
    jsgtkobj_registerWrapper(cx, gtkobj, obj);
 out:
    return obj;
}
/* end exported functions */

static void obj_signalDestroy(gpointer data) {
    JS_free(((gtkobj_signalArg *)data)->cx, data);
}

static void obj_signalHandler(GtkObject *object, gpointer data,
			      gint nparams, GtkArg *args) {
    gtkobj_signalArg *arg = (gtkobj_signalArg *)data;
    jsval *argv, rval;
    JSObject *this;
    JSBool allocated;
    uintN i;

    argv = JS_malloc(arg->cx, sizeof(jsval) * nparams);
    this = jsgtkobj_findWrapper(arg->cx, GTK_OBJECT(object));
    if (!this) {
	printf("WARNING: No wrapper found for %x, so creating one\n",
	       object);
	this = jsgtkobj_wrapObject(arg->cx, object, NULL);
	if (!this)
	    goto out;
	JS_AddRoot(arg->cx, this);
	allocated = JS_TRUE;
    }
    for (i = 0; i < nparams; i++) {
	JS_AddRoot(arg->cx, &argv[i]);
	switch(args[i].type) {
	case GTK_TYPE_OBJECT:
	    printf("arg %d/%s is object: %x\n", i, args[i].name, args[i]);
	    argv[i] = OBJECT_TO_JSVAL(jsgtkobj_findWrapper(arg->cx, 
						  GTK_VALUE_OBJECT(args[i])));
	    if (!argv[i]) {
		printf("WARNING: No wrapper found for %x, so creating one\n",
		       args[i]);
		argv[i] = OBJECT_TO_JSVAL(jsgtkobj_wrapObject(arg->cx,
						  GTK_VALUE_OBJECT(args[i]),
							      NULL));
		if (!argv[i])
		    goto out;
	    }
	break;
	default:
	    printf("arg %d/%s is unknown type %s\n", i, args[i].name,
		   gtk_type_name(args[i].type));
	    break;
	}
    }
    JS_CallFunctionValue(arg->cx, this, OBJECT_TO_JSVAL(arg->obj),
			 nparams, argv, &rval);

 out:
    if (allocated)
	JS_RemoveRoot(arg->cx, this);
    for (i = 0; i < nparams; i++)
	JS_RemoveRoot(arg->cx, &argv[i]);
    JS_free(arg->cx, argv);
}

static JSBool obj_disconnect(JSContext *cx, JSObject *obj,
			  uintN argc, jsval *argv, jsval *rval) {
    GtkObject *gtkobj;

    if (!argc || !JSVAL_IS_INT(argv[0]))
	return JS_FALSE;
    gtkobj = JS_GetPrivate(cx, obj);
    if (!GTK_IS_OBJECT(gtkobj))
	return JS_FALSE;
    gtk_signal_disconnect(gtkobj, JSVAL_TO_INT(argv[0]));
    return JS_FALSE;
}

static JSBool obj_connect(JSContext *cx, JSObject *obj,
			  uintN argc, jsval *argv, jsval *rval) {
    gtkobj_signalArg *arg;
    GtkObject *gtkobj;
    gint id;

    if (argc < 2 ||
	!JSVAL_IS_STRING(argv[0]) ||
	!JSVAL_IS_OBJECT(argv[1]) ||
	!JS_ValueToFunction(cx, argv[1]))
               return JS_FALSE;
    if (argc == 3)
      {
	if (!JSVAL_IS_STRING(argv[2]))
	  return JS_FALSE;
      }
    gtkobj = JS_GetPrivate(cx, obj);
    if (!GTK_IS_OBJECT(gtkobj))
	return JS_FALSE;
    /* check for existing arg before we overwrite it? */
    arg = JS_malloc(cx, sizeof(gtkobj_signalArg));
    arg->cx = cx;
    arg->obj = JSVAL_TO_OBJECT(argv[1]);

    id = gtk_signal_connect_full(
			    gtkobj, 
			    JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),
			    NULL,
			    obj_signalHandler,
			    (gpointer) arg,
			    obj_signalDestroy,
			    0,
			    0
			    ); 
/*
gint gtk_signal_connect_interp (GtkObject *object, gchar *name, GtkCallbackMarshal func,
gpointer data, GtkDestroyNotify destroy_func, gint after) 
*/
    /*
      id = gtk_signal_connect_interp(gtkobj, 
      JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),
      obj_signalHandler, arg,
      obj_signalDestroy, 0);
    */
    
    /* failure? */
    *rval = INT_TO_JSVAL(id);
    return JS_TRUE;
}

static JSFunctionSpec object_methods[] = {
    {"connect", 		obj_connect,		2},
    {"disconnect",		obj_disconnect,		1},
    {0}
};

/* From other js-gtk bits */
JSBool jsgtkwidget_init(JSContext *cx, JSObject *global, JSObject *gtkproto);

struct JSConstIntSpec {
    jsint	ival;
    const char	*name;
    uint8	flags;
};

#define INT_ENUM(name) \
{GTK_##name, #name, JSPROP_READONLY|JSPROP_PERMANENT|JSPROP_ENUMERATE }

struct JSConstIntSpec GTKprops[] = {
    INT_ENUM(STATE_NORMAL),
    INT_ENUM(STATE_ACTIVE),
    INT_ENUM(STATE_PRELIGHT),
    INT_ENUM(STATE_SELECTED),
    INT_ENUM(STATE_INSENSITIVE),
    INT_ENUM(WINDOW_TOPLEVEL),
    INT_ENUM(WINDOW_DIALOG),
    INT_ENUM(WINDOW_POPUP),
    INT_ENUM(DIR_TAB_FORWARD),
    INT_ENUM(DIR_TAB_BACKWARD),
    INT_ENUM(DIR_UP),
    INT_ENUM(DIR_DOWN),
    INT_ENUM(DIR_LEFT),
    INT_ENUM(DIR_RIGHT),
    INT_ENUM(SHADOW_NONE),
    INT_ENUM(SHADOW_IN),
    INT_ENUM(SHADOW_OUT),
    INT_ENUM(SHADOW_ETCHED_IN),
    INT_ENUM(SHADOW_ETCHED_OUT),
    INT_ENUM(ARROW_UP),
    INT_ENUM(ARROW_DOWN),
    INT_ENUM(ARROW_LEFT),
    INT_ENUM(ARROW_RIGHT),
    INT_ENUM(PACK_START),
    INT_ENUM(PACK_END),
    INT_ENUM(POLICY_ALWAYS),
    INT_ENUM(POLICY_AUTOMATIC),
    INT_ENUM(UPDATE_CONTINUOUS),
    INT_ENUM(UPDATE_DISCONTINUOUS),
    INT_ENUM(UPDATE_DELAYED),
    INT_ENUM(EXPAND),
    INT_ENUM(SHRINK),
    INT_ENUM(FILL),
    INT_ENUM(RUN_FIRST),
    INT_ENUM(RUN_LAST),
    INT_ENUM(RUN_BOTH),
/*    INT_ENUM(RUN_MASK),*/
    INT_ENUM(RUN_NO_RECURSE),
    INT_ENUM(WIN_POS_NONE),
    INT_ENUM(WIN_POS_CENTER),
    INT_ENUM(WIN_POS_MOUSE),
    INT_ENUM(DIRECTION_LEFT),
    INT_ENUM(DIRECTION_RIGHT),
    INT_ENUM(TOP_BOTTOM),
    INT_ENUM(LEFT_RIGHT),
    INT_ENUM(MENU_FACTORY_MENU),
    INT_ENUM(MENU_FACTORY_MENU_BAR),
    INT_ENUM(MENU_FACTORY_OPTION_MENU),
    INT_ENUM(PIXELS),
    INT_ENUM(INCHES),
    INT_ENUM(CENTIMETERS),
    INT_ENUM(SCROLL_NONE),
    INT_ENUM(SCROLL_STEP_BACKWARD),
    INT_ENUM(SCROLL_STEP_FORWARD),
    INT_ENUM(SCROLL_PAGE_BACKWARD),
    INT_ENUM(SCROLL_PAGE_FORWARD),
    INT_ENUM(TROUGH_NONE),
    INT_ENUM(TROUGH_START),
    INT_ENUM(TROUGH_END),
    INT_ENUM(POS_LEFT),
    INT_ENUM(POS_RIGHT),
    INT_ENUM(POS_TOP),
    INT_ENUM(POS_BOTTOM),
    INT_ENUM(PREVIEW_COLOR),
    INT_ENUM(PREVIEW_GRAYSCALE),
    INT_ENUM(JUSTIFY_LEFT),
    INT_ENUM(JUSTIFY_RIGHT),
    INT_ENUM(JUSTIFY_CENTER),
    INT_ENUM(JUSTIFY_FILL),
    {0}
};

static JSBool gtkobj_GTKinit(JSContext *cx, JSObject *global) {
    JSObject *GTK_obj;
    struct JSConstIntSpec *iter;
    jsval rval;

    if (!JS_EvaluateScript(cx, global, "new Object()", 12,
			   NULL, 0, &rval))
	return JS_FALSE;
    
    GTK_obj = JSVAL_TO_OBJECT(rval);
    if (!JS_DefineProperty(cx, global, "GTK", rval, NULL, NULL, 
			   JSPROP_READONLY|JSPROP_PERMANENT))
	return JS_FALSE;
    for (iter = GTKprops; iter->name; iter++)
	if (!JS_DefineProperty(cx, GTK_obj, iter->name,
			       INT_TO_JSVAL(iter->ival), NULL, NULL,
			       iter->flags))
	    return JS_FALSE;
    return JS_TRUE;
}


JSBool jsgtkobj_init(JSContext *cx, JSObject *obj, int argc, char **argv) {
    JSBool ok = JS_FALSE;
    JSObject *GTK_proto;

    /* init the wrapper hash table */
    /* wrapperHash = PR_NewHashTable(32, StubHash, PR_CompareValues,
    		  PR_CompareValues, NULL, NULL);
    */
    /* JG 1999 02 27 */
    wrapperHash = g_hash_table_new(g_direct_hash, g_direct_equal);
    if (!wrapperHash)
	return JS_FALSE;

    /* Classes, constructors and prototypes, etc. */
    GTK_proto = JS_InitClass(cx, obj, NULL, &jsgtk_ObjectClass,
			     NULL, 0,
			     NULL, object_methods,
			     NULL, NULL);
    if (!GTK_proto)
	goto out;
    
    /* add 'GTK' to scope and attendant bits */
    if (!gtkobj_GTKinit(cx, obj))
	goto out;

    /* init child widgets */
    if (!jsgtkwidget_init(cx, obj, GTK_proto))
	goto out;
    ok = JS_TRUE;
 out:
    if (!ok)
	/* JG 1999 02 27 */
	/* PR_HashTableDestroy(wrapperHash);*/
	g_hash_table_destroy(wrapperHash);

    gtk_init(&argc, (char ***)&argv);
    printf("GTK initialized\n");
    return ok;
}

